/* digjoy.c ********************************* updated: 1998-02-20 15:31 TT
 *
 * Author      : Tormod Tjaberg
 * Created     : 1997-05-08
 * Description : Routines for testing the digital joystick interface
 *
 * Changes:
 * 1998-02-19: 1.10 : Now displays available ports + bugfix forgot the %d
 *                    when printing out the status for a one button joystick.
 */
#include <stdlib.h>
#include <stdio.h>
#if defined(__DJGPP__)
#include <sys/farptr.h>
#include <go32.h>
#endif

#define PROGRAM_VER  "1.10"

volatile int joy_left, joy_right, joy_up, joy_down;
volatile int joy_b1;
volatile int joy_b2;
volatile int joy_b3;
volatile int joy_b4;

static unsigned short DataRegAddr = 0;
static unsigned short StatusRegAddr = 0;
static unsigned short CtrlRegAddr = 0;
static unsigned char  OldCtrlRegVal = 0;
static unsigned char  OldDataRegVal = 0;

#if defined(__DJGPP__)
#include <sys/farptr.h>
#include <go32.h>
#else
/* Fake things a little bit for the other compilers */
#define _dos_ds   0
#include <dos.h>

unsigned short _farpeekw(unsigned short Selector, unsigned long Offset)
{
#if defined(__386__)  /* Watcom DOS4GW */
   return *((unsigned short *) Offset);
#elif defined(DOS386) /* Symantec, Zortech, DOSX, PHARLAP */
   return *((unsigned short far *) MK_FP(_x386_zero_base_selector, Offset));
#else
   return *((unsigned short far *) Offset);
#endif
}
#endif

/* forward declarations */
short InitDigJoy(short PortNum);
void PollDigJoyFourBut(void);
void PollDigJoyOneBut(void);
void PollDigJoyNone(void);

/* 
 * Restore the registers to their initial state
 */
void TermDigJoy(void)
{
   outp(CtrlRegAddr, OldCtrlRegVal);
   outp(DataRegAddr, OldDataRegVal);
   
   DataRegAddr = 0;
   StatusRegAddr = 0;
   CtrlRegAddr = 0;
   OldCtrlRegVal = 0;
   OldDataRegVal = 0;
}

/* Input  : port number (1 - 4)
 * Returns: -1 - port number out of range 
 *          -2 - invalid printer port specified
 *          -3 - incompatible printer port
 *           0 - support for 1 fire button
 *           1 - support for 4 fire buttons
 * 
 * Init the digital joystick interface. Try to determine what
 * kind of capabilities we've got.
 */
short InitDigJoy(short PortNum)
{
   short RetVal = 0;

   if (PortNum < 1 || PortNum > 4)
      return -1;

   joy_left = 0;
   joy_right = 0;
   joy_up = 0;
   joy_down = 0;
   joy_b1 = 0; 
   joy_b2 = 0; 
   joy_b3 = 0; 
   joy_b4 = 0; 

   DataRegAddr = _farpeekw(_dos_ds, 0x40 * 16 + 0x8 + ((PortNum - 1) * 2));

   if (DataRegAddr == 0)
      return -2;

   StatusRegAddr = DataRegAddr + 1;
   CtrlRegAddr = DataRegAddr + 2;
   OldCtrlRegVal = inp(CtrlRegAddr);
   OldDataRegVal = inp(DataRegAddr);

   /* Make sure we can support at least 1 fire button */
   outp(DataRegAddr, 0xff);
   if (((inp(StatusRegAddr) >> 3) & 0x1f) != 0xf)
   {  
      TermDigJoy();
      return -3;
   }

   /* 
    * Check if we are able to control the following lines (which we use for input!)
    * . 14 (AUTO FEED)
    * . 16 (INIT)
    * . 17 (SLCT IN)
    */

   /* Turn off all data lines send 0xa to the register and read it back */
   outp(DataRegAddr, 0x0);
   outp(CtrlRegAddr, 0xa);

   if (((inp(CtrlRegAddr) >> 1) & 0x7) != 0x5)
      RetVal = 0;   /* At least we can support 1 fire button */
   else
      RetVal = 1;

   /* Turn on all data lines send 0x4 and read it back, make sure
    * irq is not enabled. Not really a problem since you have to
    * enable it using the mask register as well. Note we have to
    * turn on all data lines to enable the interface properly.
    */
   outp(DataRegAddr, 0xff);
   outp(CtrlRegAddr, 0x4);

   if (((inp(CtrlRegAddr) >> 1) & 0x7) != 0x2)
      RetVal = 0;   /* At least we can support 1 fire button */
   else
      RetVal = 1;
      
   atexit(TermDigJoy);

   return RetVal;
}  

/* 
 * Poll the parallel port setting the appropriate values.
 */
void PollDigJoyFourBut(void)
{
   unsigned short us;

   outp(CtrlRegAddr, 0x4);

   /* Get the exotic switches */
   us = inp(CtrlRegAddr);

   /* Is it faster to use a switch with 9 cases ? */
   joy_b2 = (us & 0x2) ? 1 : 0;
   joy_b3 = (us & 0x4) ? 0 : 1;
   joy_b4 = (us & 0x8) ? 1 : 0;

   us = inp(StatusRegAddr);

   joy_b1 = (us & 0x80) ? 1 : 0;

   us = (us >> 3) & 0xf;

   /* 8 actual positions available ! */
   switch(us)
   {
      case 0x7 : joy_up = 1; joy_down = 0; joy_left = 0; joy_right = 0; break;
      case 0xb : joy_up = 0; joy_down = 1; joy_left = 0; joy_right = 0; break;
      case 0xd : joy_up = 0; joy_down = 0; joy_left = 1; joy_right = 0; break;
      case 0xe : joy_up = 0; joy_down = 0; joy_left = 0; joy_right = 1; break;
      case 0x6 : joy_up = 1; joy_down = 0; joy_left = 0; joy_right = 1; break;
      case 0x5 : joy_up = 1; joy_down = 0; joy_left = 1; joy_right = 0; break;
      case 0xa : joy_up = 0; joy_down = 1; joy_left = 0; joy_right = 1; break;
      case 0x9 : joy_up = 0; joy_down = 1; joy_left = 1; joy_right = 0; break;
      default  : joy_up = 0; joy_down = 0; joy_left = 0; joy_right = 0; break;
   }
}

/* 
 * Poll the parallel port setting the appropriate values.
 * This function has only got support for 1 button. If I can't
 * get the three extra buttons there is no to read them.
 */
void PollDigJoyOneBut(void)
{
   unsigned short us;

   us = inp(StatusRegAddr);

   joy_b1 = (us & 0x80) ? 1 : 0;

   us = (us >> 3) & 0xf;

   /* 8 actual positions available ! */
   switch(us)
   {
      case 0x7 : joy_up = 1; joy_down = 0; joy_left = 0; joy_right = 0; break;
      case 0xb : joy_up = 0; joy_down = 1; joy_left = 0; joy_right = 0; break;
      case 0xd : joy_up = 0; joy_down = 0; joy_left = 1; joy_right = 0; break;
      case 0xe : joy_up = 0; joy_down = 0; joy_left = 0; joy_right = 1; break;
      case 0x6 : joy_up = 1; joy_down = 0; joy_left = 0; joy_right = 1; break;
      case 0x5 : joy_up = 1; joy_down = 0; joy_left = 1; joy_right = 0; break;
      case 0xa : joy_up = 0; joy_down = 1; joy_left = 0; joy_right = 1; break;
      case 0x9 : joy_up = 0; joy_down = 1; joy_left = 1; joy_right = 0; break;
      default  : joy_up = 0; joy_down = 0; joy_left = 0; joy_right = 0; break;
   }
}

/* 
 * Don't do anything at all. Handy if you don't want any joystick support!
 */
void PollNoJoy(void)
{
}

/* This code was adapted from the "bidir.c" code by Kris Heidenstrom
 * in PC Parallel Port Mini-FAQ revision 7
 */

/* Checks whether the port's data register retains data, returns 1 if
 * so, 0 if not.  The data register retains data on non-bidirectional
 * ports, but on bidirectional ports in high impedance (tri-state)
 * mode, the data register will not retain data. 
 */
char fTestRetention(unsigned short IO_Base) 
{
   outp(IO_Base, 0x55);                      /* Write a new value */
   (void) inp(IO_Base);                      /* Delay */
   if (inp(IO_Base) != 0x55) 
      return 0;                             /* Did not retain data */

   outp(IO_Base, 0xAA);                      /* Write another new value */
   (void) inp(IO_Base);                      /* Delay */
   if (inp(IO_Base) != 0xaa) 
      return 0;                             /* Did not retain data */
   return 1;                                /* Retained data alright */
}

/* Portnum is a number between 1 & 4 */
void DeterminePortType(short PortNum) 
{
   unsigned short IO_Base;
   unsigned short OldCtrl;
   unsigned short OldVal;

   IO_Base = _farpeekw(_dos_ds, 0x40 * 16 + 0x8 + ((PortNum - 1) * 2));

   if (IO_Base == 0) 
   {
      printf("Parallel port %d does not exist\n", PortNum);
      return;
   }

   OldCtrl = inp(IO_Base + 2);
   outp(IO_Base + 2, OldCtrl & 0xDF);        /* Bidir off */
   (void) inp(IO_Base);                      /* Delay */
   OldVal = inp(IO_Base);                    /* Keep old data */
   if (fTestRetention(IO_Base) == 0) 
   {
      printf("Parallel port %d (0x%x) is faulty or set to input mode\n", PortNum, IO_Base);
      goto ReturnAndRestore;
   }
   outp(IO_Base + 2, OldCtrl | 0x20);        /* Bidir on for some ports */
   if (fTestRetention(IO_Base))
      printf("Parallel port %d (0x%x) is non-bidirectional or in standard mode\n", PortNum, IO_Base);
   else
      printf("Parallel port %d (0x%x) is bidirectional using control port bit 5\n", PortNum, IO_Base);

ReturnAndRestore:
   outp(IO_Base + 2, OldCtrl);               /* Put it back */
   outp(IO_Base, OldVal);                    /* Restore data */
   return;
}

void FindPort(void)
{
   short PortNum;

   for (PortNum = 1; PortNum <= 4; PortNum++)
      DeterminePortType(PortNum);

   putchar('\n');
}

void ShowInfo(void)
{
   puts("This program will test the digital joystick interface. Make sure the");
   puts("joystick is centred and that the fire button(s) are not pressed before");
   puts("running this program. Read the documentation for troubleshooting tips.\n");
}

int main(int argc, char *argv[])
{
   short i;
   short s;
   void (*pPollDigJoy)(void);

   puts("digjoy, ver " PROGRAM_VER " -- Copyright (C) 1997-98 Tormod Tjaberg\n");

   if (argc == 1)
   {
      puts("Usage: digjoy <port number>\n");
      ShowInfo();
      FindPort();
      puts("command error, missing parallel port number 1 - 4");
      return 1;
   }

   ShowInfo();
   FindPort();

   s = argv[1][0] - '0';
   i = InitDigJoy(s);

   if (i < 0)
   {
      switch(i)
      {
         case -1: printf("command error, port number %d is out of range (1-4)\n", s); break;
         case -2: printf("error, parallel port %d does not exist\n", s); break;
         case -3: printf("sorry, your parallel port %d does not support the interface\n", s); break;
         default: printf("InitDigJoy returns %d\n", i); break;
      }
      return 1;
   }

   if (i == 0)
   {
      puts("Support for 1 button joystick detected");
      pPollDigJoy = PollDigJoyOneBut;
   }

   if (i == 1)
   {
      puts("Support for 4 button joystick detected");
      pPollDigJoy = PollDigJoyFourBut;
   }

   printf("DataRegAddr   = 0x%x (0x%x)\n", DataRegAddr, inp(DataRegAddr));
   printf("StatusRegAddr = 0x%x (0x%x)\n", StatusRegAddr, inp(StatusRegAddr));
   printf("CtrlRegAddr   = 0x%x (0x%x)\n", CtrlRegAddr, inp(CtrlRegAddr));

   printf("\nTest the joystick functions. Hit any key to exit program.\n\n");
   while (!kbhit())
   {
      pPollDigJoy();

      if (i == 1)
         printf("UP %d DOWN %d LEFT %d RIGHT %d FIRE_1 %d FIRE_2 %d FIRE_3 %d FIRE_4 %d\r", 
                joy_up, joy_down, joy_left, joy_right, joy_b1, joy_b2, joy_b3, joy_b4);  
      else
         printf("UP %d DOWN %d LEFT %d RIGHT %d FIRE %d\r", 
                joy_up, joy_down, joy_left, joy_right, joy_b1);
   }

   putchar('\n');

   return 0;
}
